"""
Copyright (c) 2012-2021 (https://github.com/clippercard/clippercard-python)

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

# === imports ===

import bs4
import requests
import clippercard.parser as parser


# === Error Classes ===


class ClipperCardError(Exception):
    """base error for client"""


class ClipperCardAuthError(ClipperCardError):
    """unable to login with provided credentials"""


class ClipperCardContentError(ClipperCardError):
    """unable to recognize and parse web content"""


# === ClipperCardWebSession ===


class ClipperCardWebSession(requests.Session):
    """
    A stateful session for clippercard.com
    """

    LOGIN_URL = "https://www.clippercard.com/ClipperWeb/account"
    BALANCE_URL = "https://www.clippercard.com/ClipperWeb/account.html"
    HEADERS = {
        "FAKE_USER_AGENT": (
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/90.0.4430.85 Safari/537.36"
        )
    }

    def __init__(self, username=None, password=None):
        requests.Session.__init__(self)
        self.headers.update(self.HEADERS)
        self._account_resp_soup = None
        if username and password:
            self.login(username, password)

    def login(self, username, password):
        """
        Mimicks user login form submission
        """
        login_landing_resp = self.get(self.LOGIN_URL)

        if login_landing_resp.ok:
            req_data = {
                "_csrf": parser.parse_login_form_csrf(login_landing_resp.text),
                "email": username,
                "password": password,
            }
        else:
            raise ClipperCardError(
                "ClipperCard.com site may be down right now. Try again later."
            )

        login_resp = self.post(self.LOGIN_URL, data=req_data)
        if not login_resp.ok:
            raise ClipperCardError(
                "ClipperCard.com site may be down right now. Try again later."
            )

        resp_soup = bs4.BeautifulSoup(login_resp.text, "html.parser")
        possible_error_msg = resp_soup.find(
            "div", attrs={"class": "form-error-message"}
        )
        if possible_error_msg is not None:
            raise ClipperCardAuthError(
                parser.cleanup_whitespace(possible_error_msg.get_text())
            )

        # assume account page is reachable now
        self._account_resp_soup = resp_soup
        return login_resp

    @property
    def profile_info(self):
        """
        Returns *Profile* namedtuples associated with logged in user
        """
        if not self._account_resp_soup:
            raise ClipperCardError("Must login first")
        return parser.parse_profile_info(self._account_resp_soup)

    @property
    def cards(self):
        """
        Returns list of *Card* namedtuples associated with logged in user
        """
        if not self._account_resp_soup:
            raise ClipperCardError("Must login first")
        return parser.parse_cards(self._account_resp_soup)

    def print_summary(self):
        """return a text summary of the account"""
        print(self.profile_info)
        print("=" * 80)
        for card in sorted(self.cards, key=lambda card: card.serial_number):
            print(card)
            print("-" * 80)



"""
ClipperCard Client

Usage:
  clippercard (-h | --version)
  clippercard summary {auth_args}

Options:
  -h --help             Show this screen.
  -v --version          Show version.
  --account=<acct>      Optional, account section name in ~/.clippercardrc config [Default: default].
  --config=<cfg>        Optional, account login config file path [Default: ~/.clippercardrc].
  --username=<user>     Optional, login username if not using config file.
  --password=<pass>     Optional, login password if not using config file.
"""

__doc__ = __doc__.format(
    auth_args="([--account=<acct>] [--config=<cfg>] | [ (--username=<user> --password=<pass>) ])"
)

import configparser
import logging
import os
import sys

import docopt
import clippercard
import clippercard.porcelain


class ClipperCardCommandError(Exception):
    pass


def _get_client_auth(args):
    """
    Finds/parses the username and password from either the args or a config file.

    :returns: a tuple of (username, password)
    """
    username, password = args["--username"], args["--password"]
    if not (username and password):
        config_file_path = os.path.expanduser(args["--config"])
        if not os.path.exists(config_file_path):
            raise ClipperCardCommandError(
                "Login config file {0} does not exist.".format(config_file_path)
            )
        try:
            parser = configparser.SafeConfigParser()
            parser.read(config_file_path)
            section = args["--account"]
            username, password = parser.get(section, "username"), parser.get(
                section, "password"
            )
        except configparser.NoSectionError:
            raise ClipperCardCommandError(
                "Account config section {0} is not found in {1}".format(
                    repr(args["--account"]), config_file_path
                )
            )
    return username, password


def main():
    args = docopt.docopt(__doc__, version=clippercard.__version__)
    if args["--version"]:
        print(clippercard.__version__)
        sys.exit(0)

    try:
        username, password = _get_client_auth(args)
        session = clippercard.Session(username, password)
        if args["summary"]:
            print(
                clippercard.porcelain.tabular_output(
                    session.profile_info, session.cards
                )
            )
    except (clippercard.client.ClipperCardError, ClipperCardCommandError) as e:
        sys.exit(str(e))


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()



"""
Copyright (c) 2012-2021 (https://github.com/clippercard/clippercard-python)

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import collections
import itertools
import logging
import re
import bs4


logger = logging.getLogger(__name__)


# === Simple Objects for ClipperCard data ===

_profile_fields = [
    "name",
    "email",
    "mailing_address",
    "phone",
    "alt_phone",
    "primary_payment",
    "backup_payment",
]


class Profile(collections.namedtuple("Profile", _profile_fields)):
    """a simple class to represent a user profile"""

    def __str__(self):
        return "\n".join(
            [
                "Name: {name}",
                "Email: {email}",
                "Address: {mailing_address}",
                "Primary Phone Number: {phone}",
                "Alternate Phone Number: {alt_phone}",
                "Primary Payment: {primary_payment}",
                "Backup Payment: {backup_payment}",
            ]
        ).format(**self._asdict())


_product_fields = ["name", "value"]  # e.g. Cash value, BART HVD 60/64


class CardFeature(collections.namedtuple("CardFeature", _product_fields)):
    """a simple class to represent a card feature e.g. reload plan"""

    def __str__(self):
        return "{name}: {value}".format(**self._asdict())


class CardProduct(collections.namedtuple("CardProduct", _product_fields)):
    """a simple class to represent a card product e.g. cash value or pass"""

    def __str__(self):
        return "{name}: {value}".format(**self._asdict())


_card_fields = [
    "serial_number",
    "nickname",
    "type",  # Adult, Senior, Youth, Disabled Discount
    "status",  # Active, Inactive
    "features",  # a list of annotated properties of the card (e.g. auto-load)
    "products",  # a list of CardProduct, e.g. Cash value, Train Pass
]


class Card(collections.namedtuple("Card", _card_fields)):
    """a simple class to represent an instance of Clipper Card"""

    def __str__(self):
        lines = [
            '{serial_number} "{nickname}" ({type} - {status})'.format(**self._asdict())
        ]
        for prod in self.products:
            lines.append(f"  - {prod}")
        for feat in self.features:
            lines.append(f"  - {feat}")
        return "\n".join(lines)


# === Helpers ===

REGEX_WHITESPACE = re.compile(r"\s+")  # noqa


def cleanup_whitespace(text_content):
    """clean up junk whitespace that comes with every table cell"""
    return re.sub(REGEX_WHITESPACE, " ", text_content.strip())


# === Section Parsers ===


def parse_login_form_csrf(login_page_content):
    """Parse the login form the _csrf arg for login submission"""
    soup = bs4.BeautifulSoup(login_page_content, "html.parser")
    login_form = soup.find("form", id="login-form")
    csrf_value = login_form.find("input", attrs={"name": "_csrf"}).attrs["value"]
    return csrf_value


def parse_cards(account_page_soup):
    """Parse the list of Clipper Cards registered to the profile"""
    card_info_divs = account_page_soup.find_all(
        "div", attrs={"class": "clipper-card-info", "data-parent": "#clipper-cards"}
    )
    card_names = []
    card_name_headers = account_page_soup.find_all(
        "h2", attrs={"class": "clipper-card-name"}
    )
    for card_name_h2 in card_name_headers:
        card_names.append(
            card_name_h2.find("span", attrs={"class": "sr-only"}).get_text()
        )
    cards = []
    for i, info_div in enumerate(card_info_divs):
        card_id = info_div.attrs["id"].replace("clipper-card-info-", "")
        big_money_value = info_div.find(
            "p",
            attrs={
                "class": "big-money",
            },
        )
        products = [
            CardProduct(
                name="Cash Value", value=big_money_value.get_text().replace(" ", "")
            )
        ]
        features = []
        feature_bullets = info_div.find("ul", attrs={"class": "bullets"})
        if feature_bullets is not None:
            for bullet_item in feature_bullets.find_all("li"):
                features.append(
                    CardFeature(name="Reload", value=bullet_item.get_text())
                )
        current_passes_div = info_div.find(
            "div", attrs={"class": "current-passes-section"}
        )
        card_type = None
        card_status = None
        all_pass_info = current_passes_div.find_all("p") or []
        for pass_info in all_pass_info:
            try:
                if "Current Passes" in pass_info.get_text():
                    products.append(
                        CardProduct(
                            name="Current Passes",
                            value=cleanup_whitespace(
                                pass_info.find_next_sibling().get_text()
                            ),
                        )
                    )
                elif "Pending Passes" in pass_info.get_text():
                    products.append(
                        CardProduct(
                            name="Pending Passes",
                            value=cleanup_whitespace(
                                pass_info.find_next_sibling().get_text()
                            ),
                        )
                    )
                elif "Card Type" in pass_info.get_text():
                    card_type = cleanup_whitespace(
                        pass_info.find_all("span")[-1].get_text()
                    )
                elif "Card Status" in pass_info.get_text():
                    card_status = cleanup_whitespace(
                        pass_info.find_all("span")[-1].get_text()
                    )
            except Exception as err:  # noqa
                logger.error("Parse error on pass_info: {%s}\n{%s}", err, pass_info)

        cards.append(
            Card(
                serial_number=card_id,
                nickname=card_names[i],
                type=card_type,
                status=card_status,
                features=features,
                products=products,
            )
        )
    return cards


def parse_profile_info(account_soup):
    """Parse the attributes of the logged in user"""
    profile_info_div = account_soup.find("div", attrs={"id": "profile-info"})
    expected_data_items = [
        "name",
        "email",
        "_",  # ignore, email_updates_enabled
        "mailing_address",
        "_",  # ignore, primary_phone_label
        "phone",
        "_",  # ignore, alt_phone_label
        "alt_phone",
    ]
    profile_data_spans = profile_info_div.find_all("span")
    profile_info = {}
    for data_item, data_span in itertools.zip_longest(
        expected_data_items, profile_data_spans
    ):
        if data_item != "mailing_address":
            profile_info[data_item] = cleanup_whitespace(data_span.get_text())
        else:
            profile_info[data_item] = cleanup_whitespace(
                data_span.find_parent().get_text().replace("Mailing Address", "")
            )
    profile_info.pop("_")
    profile_info["primary_payment"] = (
        account_soup.find("div", attrs={"id": "payment-info"})
        .find_all("span")[-1]
        .get_text()
    )
    profile_info["backup_payment"] = (
        account_soup.find("div", attrs={"id": "backup-payment-info"})
        .find_all("span")[-1]
        .get_text()
    )
    return Profile(**profile_info)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name="clippercard",
    version="202104.25.2",  # year/month, day, release of day
    author="Your friendly neighborhood transit rider-hackers",
    author_email="goldengate88@systemfu.com",
    packages=["clippercard", "tests"],
    package_dir={"clippercard": "clippercard"},
    entry_points={
        "console_scripts": [
            "clippercard = clippercard.main:main",
        ]
    },
    scripts=[],
    url="https://github.com/clippercard/clippercard-python",
    license="MIT",
    description="Unofficial Python API for Clipper Card (transportation pass used in the SF Bay Area)",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "BeautifulSoup4>=4.9.0",
        "docopt>=0.6.1,<1.0",
        "PrettyTable>=2.1.0",
        "requests>=2.25.0,<3.0",
        "urllib3>=1.26.0"
    ],
    tests_require=[
        "pytest>=6.2.0",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
